#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Unit conversion calculators.

CHANGELOG:
  2005/07/14 CB: Initial import.
  2005/07/15 MO: Latitude and Longitude classes.
  2005/07/18 MO: Limit Latitude range to 90.
  2007/05/03 CB: Tweaked lat-long to get correct answer for 40 minutes.
                 Added "convert" as alias for "Convert"
  2007/12/17 MO: Add .format() method to Latitude/Longitude returning Unicode.
                 .factor_html() remains for backward compatibility, using
                 .format() internally.
                 Make Latitude/Longitude.__repr__ more robust in case
                 .__init__ raises an exception; workaround for Pylons bug
                 http://pylonshq.com/project/pylonshq/ticket/341
  2008/02/22 CB: Added a few more units for Ian
  2008/06/05 CB: Various changes before putting the Converter GUI on the web:
                 new units, changed spelling, etc.
  2009/09/29 CB: Re-factored the lat-long stuff:
                 - it's not in a separate module
                 - Mike and Chris' code has been merged for less duplication
                 - Unit data moved to separate module
  2018/01/24 CB: Added unicode exponents in unit names
                 Added __all__
                 Fixed concentration in water units!
"""

import warnings

from .unit_data import ConvertDataUnits

from nucos import lat_long

from .lat_long import (LatLongConverter,
                       Latitude,
                       Longitude,
                       format_lat,
                       format_lon,
                       format_lat_d,
                       format_lon_d,
                       format_lat_dm,
                       format_lon_dm,
                       format_lat_dms,
                       format_lon_dms,
                       )

# reset this to get some more print statements on import
DEBUG = False

__all__ = ['DensityConverterClass',
           'InvalidUnitTypeError',
           'Latitude',
           'ConverterClass',
           'is_supported',
           'Converters',
           'GetUnitNames',
           'NotSupportedUnitError',
           'UNIT_TYPES',
           'TempConverterClass',
           'Simplify',
           'GetUnitAbbreviation',
           'Convert',
           'FindUnitTypes',
           'lat_long',
           'LatLongConverter',
           'Longitude',
           'ConvertDataUnits',
           'InvalidUnitError',
           'is_same_unit',
           'GetUnitTypes',
           'OilQuantityConverter',
           'convert',
           'UnitConversionError',
           'format_lat',
           'format_lon',
           'format_lat_dms',
           'format_lon_dms',
           ]


# A few utilities
def Simplify(string):
    """
    Simplify(string)

    returns the string with the whitespace and capitalization removed
    """
    try:
        return "".join(string.lower().split()).replace(".", "")
    except AttributeError:
        raise NotSupportedUnitError(string)


def GetUnitTypes():
    """
    returns a list of all the unit types available

    a unit type is something like "mass", "velocity", etc.
    """
    return list(ConvertDataUnits.keys())


def GetUnitNames(UnitType):
    """
    returns a list of all the units available for a given unit type available

    a unit type is something like "Mass", "Velocity", etc.

    a unit of mass would be "kilogram", "slug", etc.
    """
    return list(ConvertDataUnits[UnitType.title()].keys())



def FindUnitTypes():
    """
    Returns a mapping of all the unit names to the unit types

    Raises an exception if there is more than one option -- this will check
    the unit database for duplicated names

    Primarily used to do conversions without apecifying the unit types

    Usually not called from user code.
    """

    # compute units that are unique to Mass and Volume Fraction
    mf = ConvertDataUnits["Mass Fraction"]
    vf = ConvertDataUnits["Volume Fraction"]
    # not_to_skip = {Simplify(u) for u in mf.keys() ^ vf.keys()}
    to_skip = {Simplify(u) for u in mf.keys() & vf.keys()}

    unit_types = {}

    for unit_type, unit_data in ConvertDataUnits.items():
        unit_type = Simplify(unit_type)

        # - skipping Oil Concentration, 'cause this is really length
        #   - lots of duplicate units!
        # - skipping Concentration in water, 'cause it's weird
        #   - mass/volume and mass/mass !
        # - skipping Mass Fraction, because there are lots of duplicate units
        #   that conflict with Concentration & Concentration In Water.
        if unit_type in ('oilconcentration',
                         'concentrationinwater',
                         #'massfraction',
                         #'volumefraction',
                         'deltatemperature',
                         'dimensionless',
                         ):
            continue

        for pname, data in unit_data.items():
            # strip out whitespace and capitalization
            pname = Simplify(pname)

            if (unit_type in {'massfraction', 'volumefraction'}
                    and pname in to_skip):
                continue

            # add the primary name:
            if pname in unit_types:
                raise ValueError(f"Duplicate primary name in units table: {pname}")
            unit_types[pname] = unit_type
            # now the synonyms:
            for n in data[1]:
                n = Simplify(n)

                # skip duplicate units, "oz" is only mass, "s" is only time
                if (unit_type, n) in [("volume", "oz"),
                                      ("density", "s")]:
                    continue
                if DEBUG:
                    try:
                        print(f"Adding: {unit_type}: {n}")
                    except UnicodeEncodeError:
                        print(f"Adding: {unit_type}: {n}".encode("utf-8"))
                if n in unit_types:
                    raise ValueError("Duplicate name in units table: %s" % n)

                unit_types[n] = unit_type

    return unit_types


def FindAllUnitNames():
    """
    returns a mapping of all the valid unit names to unit type

    Usually not called from user code.
    """

    # # compute units that are unique to Mass and Volume Fraction
    # mf = ConvertDataUnits["Mass Fraction"]
    # vf = ConvertDataUnits["Volume Fraction"]
    # # not_to_skip = {Simplify(u) for u in mf.keys() ^ vf.keys()}
    # to_skip = {Simplify(u) for u in mf.keys() & vf.keys()}

    unit_names = {}

    for unit_type, unit_data in ConvertDataUnits.items():
        unit_type = Simplify(unit_type)

        # # - skipping Oil Concentration, 'cause this is really length
        # #   - lots of duplicate units!
        # # - skipping Concentration in water, 'cause it's weird
        # #   - mass/volume and mass/mass !
        # # - skipping Mass Fraction, because there are lots of duplicate units
        # #   that conflict with Concentration & Concentration In Water.
        # if unit_type in ('oilconcentration',
        #                  'concentrationinwater',
        #                  #'massfraction',
        #                  #'volumefraction',
        #                  'deltatemperature',
        #                  'dimensionless',
        #                  ):
        #     continue

        for pname, data in unit_data.items():
            # if (unit_type in {'massfraction', 'volumefraction'}
            #         and pname in to_skip):
            #     continue

            # add the primary name:
            unit_names.setdefault(unit_type, [pname])
            # now the synonyms:
            unit_names[unit_type].extend(data[1])

    return unit_names


UNIT_TYPES = FindUnitTypes()
UNIT_NAMES = FindAllUnitNames()


def is_supported_unit(unit_type, unit):
    """
    checks if a unit name is supported for the given unit type
    """
    unit_type = Simplify(unit_type)
    converter = Converters[unit_type]
    return Simplify(unit) in converter.Synonyms


def get_supported_names(unit_type):
    """
    returns the list of all supported unit names for a unit_type in standard
    form
    """
    return UNIT_NAMES[Simplify(unit_type)]


def GetUnitAbbreviation(unit_type, unit):
    """
    return the standard abbreviation for a given unit

    :param unit_type: the type of unit: "mass", "length", etc.
    :param unit: the unit you want the abbreviation for: "gram", etc.
    """
    return ConvertDataUnits[unit_type][unit][1][0]


def is_same_unit(unit1, unit2):
    """
    Checks if the two unit names passed in are the same

    :param unit1: name of unit to compare
    :type unit1: string

    :param unit2: name of unit to compare
    :type unit2: string

    :returns: True if they are synonyms for the same unit.
              False if they are different units.
              False if one of them is not in the database.
    """
    all_types = UNIT_TYPES
    unit1 = Simplify(unit1)
    unit2 = Simplify(unit2)

    try:
        type1 = all_types[unit1]
        type2 = all_types[unit2]
    except KeyError:
        return False

    if type1 != type2:
        return False
    else:
        Synonyms = Converters[Simplify(type1)].Synonyms
        return Synonyms[Simplify(unit1)] == Synonyms[Simplify(unit2)]


class ConverterClass:
    """
    Main class for performing the conversion.
    There will be one instance for each unit type.

    sub-classes will handle special cases
    """
    def __init__(self, TypeName, UnitsDict):
        """
        Create a Converter

        :param TypeName: the name of the unit type, such as "length"
        :param UnitsDict: A dict will the unit data.
                          See unit_data.py for format
        """
        self.Name = TypeName

        self.Synonyms = {}
        self.Convertdata = {}

        for PrimaryName, data in UnitsDict.items():
            # strip out whitespace and capitalization
            Pname = Simplify(PrimaryName)

            self.Convertdata[Pname] = data[0]
            self.Synonyms[Pname] = Pname

            for synonym in data[1]:
                # duplicate check
                if synonym in self.Synonyms:
                    raise ValueError("Duplicate synonym: "
                                     "unit_type: {}, name: {}".format(TypeName,
                                                                      synonym)
                                     )
                self.Synonyms[Simplify(synonym)] = Pname

    def Convert(self, FromUnit, ToUnit, Value):
        """
        Convert(FromUnit, ToUnit, Value)

        returns a new value, in the units of ToUnit.

        :param FromUnit: the unit the original value is in
        :param ToUnit: the unit you want the value converted to
        :param Value: the original value
        """
        FromUnit = Simplify(FromUnit)
        ToUnit = Simplify(ToUnit)

        try:
            FromUnit = self.Synonyms[FromUnit]
        except KeyError:
            raise InvalidUnitError((FromUnit, self.Name))

        try:
            ToUnit = self.Synonyms[ToUnit]
        except KeyError:
            raise InvalidUnitError((ToUnit, self.Name))

        return Value * self.Convertdata[FromUnit] / self.Convertdata[ToUnit]


# the special case classes:
class TempConverterClass(ConverterClass):
    """
    Special case class for temperature conversion.

    handles the zero-offset shift for K, C, F...
    """
    def Convert(self, FromUnit, ToUnit, Value):
        """
        Convert(FromUnit, ToUnit, Value)

        returns a new value, in the units of ToUnit.

        :param FromUnit: the unit the original value is in
        :param ToUnit: the unit you want the value converted to
        :param Value: the original value
        """
        FromUnit = Simplify(FromUnit)
        ToUnit = Simplify(ToUnit)

        try:
            FromUnit = self.Synonyms[FromUnit]
        except KeyError:
            raise InvalidUnitError((FromUnit, self.Name))

        try:
            ToUnit = self.Synonyms[ToUnit]
        except KeyError:
            raise InvalidUnitError((ToUnit, self.Name))

        A1 = self.Convertdata[FromUnit][0]
        B1 = self.Convertdata[FromUnit][1]
        A2 = self.Convertdata[ToUnit][0]
        B2 = self.Convertdata[ToUnit][1]

        to_val = ((Value + B1) * A1 / A2) - B2

        return to_val


class DensityConverterClass(ConverterClass):
    """
    Special case class for Density conversion.

    handles the special case of API gravity, etc.
    """
    def Convert(self, FromUnit, ToUnit, Value):
        """
        Convert(FromUnit, ToUnit, Value)

        returns a new value, in the units of ToUnit.

        :param FromUnit: the unit the original value is in
        :param ToUnit: the unit you want the value converted to
        :param Value: the original value
        """
        FromUnit = Simplify(FromUnit)
        ToUnit = Simplify(ToUnit)

        try:
            FromUnit = self.Synonyms[FromUnit]
        except KeyError:
            raise InvalidUnitError((FromUnit, self.Name))

        try:
            ToUnit = self.Synonyms[ToUnit]
        except KeyError:
            raise InvalidUnitError((ToUnit, self.Name))

        if FromUnit == "apidegree":
            # another Special case (could I do this the same as temp?)
            Value = 141.5 / (Value + 131.5)
            FromUnit = u"specificgravity(15\xb0c)"

        if ToUnit == "apidegree":
            ToVal = (141.5 /
                     (Value * self.Convertdata[FromUnit] /
                      self.Convertdata[u"specificgravity(15\xb0c)"]) -
                     131.5)
        else:
            ToVal = (Value *
                     self.Convertdata[FromUnit] /
                     self.Convertdata[ToUnit])

        return ToVal


class OilQuantityConverter:
    """
    class for Oil Quantity conversion -- mass to/from Volume

    requires density info as well
    """
    @classmethod
    def ToVolume(self, Mass, MassUnits, Density, DensityUnits, VolumeUnits):
        """
        Convert Oil Mass to Volume

        :param Mass: mass you want converted to volume
        :param MassUnits: unit of mass input
        :param Density: density of oil
        :param DensityUnits: units of density
        :param VolumeUnits: units of volume desired

        """
        Density = convert("Density", DensityUnits, "kg/m^3", Density)
        Mass = convert("Mass", MassUnits, "kg", Mass)

        Volume = Mass / Density
        Volume = convert("Volume", "m^3", VolumeUnits, Volume)

        return Volume

    @classmethod
    def ToMass(self, Volume, VolUnits, Density, DensityUnits, MassUnits):
        """
        Convert Oil Mass to Volume

        :param Volume: volume you want converted to mass
        :param VolumeUnits: units of volume input
        :param Density: density of oil
        :param DensityUnits: units of density
        :param MassUnits: unit of mass desired for output
        """
        Density = convert("Density", DensityUnits, "kg/m^3", Density)
        Volume = convert("Volume", VolUnits, "m^3", Volume)

        Mass = Volume * Density
        Mass = convert("Mass", "kg", MassUnits, Mass)

        return Mass


# create the converter objects
Converters = {}
for (unittype, data) in ConvertDataUnits.items():
    if unittype.lower() == 'temperature':
        Converters["temperature"] = TempConverterClass(unittype, data)
    elif unittype.lower() == 'density':
        Converters["density"] = DensityConverterClass(unittype, data)
    else:
        Converters[Simplify(unittype)] = ConverterClass(unittype, data)


def is_supported(unit):
    """
    Returns True is the unit is in the list of supported units for the
    API that does not require unit_type
    """
    return Simplify(unit) in UNIT_TYPES


def convert(unit1, unit2, value, unit_type=None):
    """
    convert(unit1, unit2, value, unit_type)

    :param unit1: the unit the original value is in
    :param unit2: the unit you want the value converted to
    :param value: the original value

    If unit_type is None, then it will look in the data
    to see if it can figure out what unit type to use.

    so you should be able to do:

    convert(unit1='meter', unit2='feet', value=32)

    NOTE: Some odd units have overlapping names, so only the more common one
          is used (oz is mass, not fluid oz, for instance).
          You can get around this by using a more precise name ('fluid oz')
          or specify the unit type.

    If you do want to specify the unit type, you can use the "old" API:

      convert(unit_type, unit1, unit2, value)

    such as:

      convert('volume', 'oz', 'cc', 25)

    :param unit_type: the type of the unit: 'mass', 'length', etc.
    :param unit1: the unit the original value is in
    :param unit2: the unit you want the value converted to
    :param value: the original value
    """
    if unit_type is None:
        # the new API: no need to specify unit type
        unit1, unit2 = (Simplify(s) for s in (unit1, unit2))

        try:
            unit_type = UNIT_TYPES[unit1]
        except KeyError:
            raise NotSupportedUnitError(unit1)

        # try:
        #     unit_type2 = UNIT_TYPES[unit2]
        # except KeyError:
        #     raise NotSupportedUnitError(unit2)

        # if unit_type != unit_type2:
        #     raise UnitConversionError("Cannot convert {0} to {1}"
        #                               .format(unit1, unit2))

        unit_type = Simplify(unit_type)
    else:
        # the old API: specify the unit type
        # re-defining the inputs:
        unit_type, unit1, unit2, value = unit1, unit2, value, unit_type
        unit_type, unit1, unit2 = (Simplify(s)
                                   for s in (unit_type, unit1, unit2))

    try:
        Converter = Converters[unit_type]
    except KeyError:
        raise InvalidUnitTypeError(unit_type)

    return Converter.Convert(unit1, unit2, value)


# so as to have the old, non-PEP8 compatible name
# This is used by TapInput (any more???)
def Convert(*args, **kwargs):
    """
    so as to have the old, non-PEP8 compatible name
    This is used by TapInput (any more???)

    for new code, use convert()
    """
    warnings.warn('"Convert()" is deprecated -- use "convert()"',
                  DeprecationWarning)

    return convert(*args, **kwargs)


# fixme: we should probably simjply get rid of these and use ValueError
class UnitConversionError(ValueError):
    """
    Exception type for unit conversion errors

    Perhaps this should be subclassed more, but at the moment,
    it doesn't do anything special
    """
    pass


class NotSupportedUnitError(UnitConversionError):
    def __init__(self, unit):
        self.unit = unit

    def __str__(self):
        return ('The unit: {} is not supported or not recognized'
                .format(self.unit))


class InvalidUnitError(UnitConversionError):
    """
    Exception raised when a unit is not in the Unit conversion database
    """
    def __init__(self, unit_unit_type):
        if isinstance(unit_unit_type, (list, tuple)):
            (unit, unit_type) = unit_unit_type[:2]

            self.unit = unit
            self.type = unit_type if unit_type else ""
        else:
            super(InvalidUnitError, self).__init__(unit_unit_type)

    def __str__(self):
        if hasattr(self, 'unit'):
            return ('The unit: {} is not in the list for Unit Type: {}'
                    .format(self.unit, self.type))
        else:
            return super(InvalidUnitError, self).__str__()


class InvalidUnitTypeError(UnitConversionError):
    """
    Exception raised when a unit is not in the Unitconversion database
    """
    def __init__(self, unitType):
        self.unitType = unitType

    def __str__(self):
        return ('The unit type: {} is not in the UnitConversion database'
                .format(self.unitType))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Assorted utilities for manipulating latitude and longitude values

"""

import math
import struct


def signbit(value):
    """
    Test whether the sign bit of the given floating-point value is
    set.  If it is set, this generally means the given value is
    negative.  However, this is not the same as comparing the value
    to C{0.0}.  For example:

    >>> NEGATIVE_ZERO < 0.0
    False

    since negative zero is numerically equal to positive zero.  But
    the sign bit of negative zero is indeed set:

    >>> signbit(NEGATIVE_ZERO)
    True
    >>> signbit(0.0)
    False

    @type  value: float
    @param value: a Python (double-precision) float value

    @rtype:  bool
    @return: C{True} if the sign bit of C{value} is set;
             C{False} if it is not set.

    signbit and doubleToRawLongBits
    are from Martin Jansche:

    http://symptotic.com/mj/code.html  (MIT license).

    This is required to capture the difference between -0.0 and 0.0, which is
    useful if someone wants to convert a latitude or longitude like:
    -0.0degrees, 34minutes to  0d34'00"S

    """
    return (doubleToRawLongBits(value) >> 63) == 1


def doubleToRawLongBits(value):
    """
    @type  value: float
    @param value: a Python (double-precision) float value

    @rtype: long
    @return: the IEEE 754 bit representation (64 bits as a long integer)
             of the given double-precision floating-point value.
    """
    # pack double into 64 bits, then unpack as long int
    return struct.unpack(b'Q', struct.pack(b'd', value))[0]


class LatLongConverter(object):
    """
    Conversion for different forms of latitude and longitude coordinates.

    Decimal Degrees; Degrees, decimal minutes; degrees, minutes, seconds

    class methods available:

    ``ToDecDeg(self, d=0, m=0, s=0, ustring=False, max=180)``

    ``ToDegMin(self, DecDegrees, ustring=False)``

    ``ToDegMinSec(self, DecDegrees, ustring=False)``

    if ``ustring`` is True, the result will be a string,
    formatted nicely with Unicode, rather than numbers.

    """

    @classmethod
    def ToDecDeg(self, d=0, m=0, s=0, ustring=False, max=180):
        """
        DecDegrees = ToDecDeg(d=0, m=0, s=0)

        converts degrees, minutes, seconds to decimal degrees
        (returned as a Float).
        """
        if m < 0 or s < 0:
            raise ValueError("Minutes and Seconds have to be positive")

        if m > 60.0 or s > 60.0:
            raise ValueError("Minutes and Seconds have to be between "
                             "-180 and 180")

        if abs(d) > max:
            raise ValueError("Degrees have to be between -180 and 180")

        if signbit(d):
            Sign = -1
            d = abs(d)
        else:
            Sign = 1

        deg_has_fract = bool(math.modf(d)[0])
        min_has_fract = bool(math.modf(m)[0])

        if deg_has_fract and (m != 0.0 or s != 0.0):
            raise ValueError("degrees cannot have fraction unless both minutes"
                             "and seconds are zero")

        if min_has_fract and s != 0.0:
            raise ValueError("minutes cannot have fraction unless seconds "
                             "are zero")

        DecDegrees = Sign * (d + m/60.0 + s/3600.0)

        if ustring:
            return u"%.6f\xb0" % (DecDegrees)
        else:
            return DecDegrees

    @classmethod
    def ToDegMin(self, DecDegrees, ustring=False):
        """
        Converts from decimal (binary float) degrees to:
          Degrees, Minutes

        If the optional parameter: "ustring" is True,
        a Unicode string is returned

        """
        if signbit(DecDegrees):
            Sign = -1
            DecDegrees = abs(DecDegrees)
        else:
            Sign = 1

        Degrees = int(DecDegrees)

        # add a tiny bit then round to avoid binary rounding issues
        DecMinutes = round((DecDegrees - Degrees + 1e-14) * 60, 10)

        if ustring:
            if Sign == 1:
                return u"%i\xb0 %.3f'" % (Degrees, DecMinutes)
            else:
                return u"-%i\xb0 %.3f'" % (Degrees, DecMinutes)
        else:
            # float to preserve -0.0
            return (Sign*float(Degrees), DecMinutes)

    @classmethod
    def ToDegMinSec(self, DecDegrees, ustring=False):

        """
        Converts from decimal (binary float) degrees to:
          Degrees, Minutes, Seconds

        If the optional parameter: "ustring" is True,
        a unicode string is returned

        """
        if signbit(DecDegrees):
            Sign = -1
            DecDegrees = abs(DecDegrees)
        else:
            Sign = 1

        Degrees = int(DecDegrees)

        # add a tiny bit to avoid rounding issues
        DecMinutes = (DecDegrees - Degrees + 1e-14) * 60

        Minutes = int(DecMinutes)
        Seconds = round(((DecMinutes - Minutes) * 60), 10)

        if ustring:
            if Sign == 1:
                return u"%i\xb0 %i' %.2f\"" % (Degrees, Minutes, Seconds)
            else:
                return u"-%i\xb0 %i' %.2f\"" % (Degrees, Minutes, Seconds)
        else:
            return (Sign * float(Degrees), Minutes, Seconds)


# These are classes used in our web apps: ResponseLink, etc.
# They provide a different interface to lat-long format conversion
class Latitude(object):
    """An object that can interpret a latitude in various formats.

       Constructor:
       Latitude(deg, min=0.0, sec=0.0, direction=None)
           - 'deg' may be between -90.0 and 90.0.
           - if 'min' is nonzero, 'deg' cannot have a fractional part.
             (This means 5 and 5.0 are acceptable but 5.1 is not.)
           - if 'sec' is nonzero, 'deg' and 'min' cannot have fractional parts.
           - 'direction' may be a string beginning with 'N' or 'S' (case
             insensitive), or None.
           - if 'direction' is not None, 'deg' cannot be negative.

       Attributes:
       .value : a float in decimal degrees.  Positive is North; negative is
           South.  (These apply to zero too; positive zero is North.)

       Methods:
       .degrees() -> (float, str)
       .degrees_minutes() -> (int, float, str)
       .degrees_minutes_seconds() -> (int, int, float, str)
       The 'str' argument is the direction: "North" or "South".

       Example:
       >>> lat1 = Latitude(-120.7625)
       >>> lat2 = Latitude(-120, 45.7500)
       >>> lat3 = Latitude(-120, 45, 45)
       >>> lat4 = Latitude(120.7625, direction='South')
       >>> lat5 = Latitude(120, 45.7500, direction='S')
       >>> lat6 = Latitude(120, 45, 45, direction='south')
       >>> (lat1.value == lat2.value == lat3.value == lat4.value ==
       ... lat5.value == lat6.value)
       True
       >>> lat1.value
       -120.7625
       >>> lat1.degrees()
       (120.7625, 'South')
       >>> lat1.degrees_minutes()
       (120, 45.750000000000171, 'South')
       >>> lat1.degrees_minutes_seconds()
       (120, 45, 45.000000000010232, 'South')
       >>> print str(lat1)
       Latitude(-120.762500)
    """
    negative_direction = "South"
    positive_direction = "North"
    min = -90.0
    max = 90.0

    def __init__(self, deg, min=0.0, sec=0.0, direction=None):
        ndir = self.negative_direction[0].upper()
        pdir = self.positive_direction[0].upper()

        if direction:
            if deg < 0.0:
                msg = "degrees cannot be negative if direction is specified"
                raise ValueError(msg)

            if direction[0].upper() == pdir:
                pass
            elif direction[0].upper() == ndir:
                deg = -deg
            else:
                msg = "direction must start with %r or %r" % (pdir, ndir)
                raise ValueError(msg)

        self.value = LatLongConverter.ToDecDeg(deg, min, sec, max=self.max)

    def direction(self):
        if self.value < 0.0:
            return self.negative_direction
        else:
            return self.positive_direction

    def degrees(self):
        deg = abs(self.value)
        return deg, self.direction()

    def degrees_minutes(self):
        deg, _min = LatLongConverter.ToDegMin(abs(self.value))

        return deg, _min, self.direction()

    def degrees_minutes_seconds(self):
        deg, _min, sec = LatLongConverter.ToDegMinSec(abs(self.value))

        return deg, _min, sec, self.direction()

    def __repr__(self):
        try:
            return "%s(%f)" % (self.__class__.__name__, self.value)
        except AttributeError:
            return "%s(uninitialized)" % self.__class__.__name__

    def format(self, style):
        """
        format(style)

        returns formatted value as Unicode string with u'\xb0' (degree symbol).

        style is one of:
        1:  decimal degrees
        2:  degrees, decimal minutes
        3:  degrees, minutes, seconds
        """

        if style == 1:
            return u'''%0.2f\xb0 %s''' % self.degrees()
        elif style == 2:
            return u'''%d\xb0 %0.2f' %s''' % self.degrees_minutes()
        elif style == 3:
            return u'''%d\xb0 %d' %0.2f" %s''' % self.degrees_minutes_seconds()
        else:
            raise ValueError("style must be 1, 2, or 3")

    def format_html(self, style):
        """
        format_html(style)

        Backward compatibility for Quixote rlink and Pylons inews.
        """
        return self.format(style).replace(u"\xb0", u"&deg;")


class Longitude(Latitude):
    """See Latitude docstring.

       Positive is East; negative is West.  Degrees must be between -180.0 and
       180.0
    """
    negative_direction = "West"
    positive_direction = "East"
    min = -180.0
    max = 180.0


class DummyLatitude(object):
    """A pseudo-Latitude whose components are None.
       Useful in building HTML forms where the value is not required.

       Note: this class may be deleted if it doesn't turn out to be useful.
    """
    value = None

    def direction(self): return None

    def degrees(self): return None, None

    def degrees_minutes(self): return None, None, None

    def degrees_minutes_seconds(self): return None, None, None, None


class DummyLongitude(DummyLatitude):
    """
       Note: this class may be deleted if it doesn't turn out to be useful.
    """
    pass


# The new simple API
# -- just methods that do what we need for ResponseLink, etc.
DEGREES = "\xb0"     # "DEGREE SIGN"
MINUTES = "\u2032"   # "PRIME"
SECONDS = "\u2033"   # "DOUBLE PRIME"


LAT_POSITIVE_DIRECTION = "North"
LAT_NEGATIVE_DIRECTION = "South"
LON_POSITIVE_DIRECTION = "East"
LON_NEGATIVE_DIRECTION = "West"

FORMAT1 = "{:.2f}\N{DEGREE SIGN} {}"
FORMAT2 = "{:.0f}\N{DEGREE SIGN} {:.2f}\N{PRIME} {}"
FORMAT3 = "{:.0f}\N{DEGREE SIGN} {:.0f}\N{PRIME} {:.2f}\N{DOUBLE PRIME} {}"


def reduce_base_60(f):
    """extract the base 60 fractional portion of a floating point number.

    i.e. minutes from degrees, seconds from minutes.
    """
    fract, whole = math.modf(f)
    # Add a tiny bit before rounding to avoid binary rounding errors.
    fract = abs(fract)
    fract = (fract + 1e-14) * 60
    fract = round(fract, 10)
    return whole, fract


def format_latlon1(f, positive_direction, negative_direction):
    direction = positive_direction if f >= 0.0 else negative_direction
    return FORMAT1.format(degrees, direction)


def format_latlon2(f, positive_direction, negative_direction):
    direction = positive_direction if f >= 0.0 else negative_direction
    degrees, minutes = reduce_base_60(f)
    degrees = abs(degrees)
    return FORMAT2.format(degrees, minutes, direction)


def format_latlon3(f, positive_direction, negative_direction):
    direction = positive_direction if f >= 0.0 else negative_direction
    degrees, minutes = reduce_base_60(f)
    minutes, seconds = reduce_base_60(minutes)
    degrees = abs(degrees)
    return FORMAT3.format(degrees, minutes, seconds, direction)


def format_lat_d(degrees):
    return format_latlon1(degrees, LAT_POSITIVE_DIRECTION, LAT_NEGATIVE_DIRECTION)


def format_lat_dm(degrees):
    """
    Format a floating point number as latitude in
    degrees, decimal minutes format.

    "North" and "South" added as appropriate

    example::
      In [3]: nucos.format_lat(-33.2)
      Out[3]: '33° 12.00′ South'

    :param degrees: degrees of latitude -- negative is South
    :type degrees: float

    :returns: string version of latitude
    """

    return format_latlon2(degrees, LAT_POSITIVE_DIRECTION, LAT_NEGATIVE_DIRECTION)


def format_lat_dms(f):
    return format_latlon3(f, LAT_POSITIVE_DIRECTION, LAT_NEGATIVE_DIRECTION)


def format_lon_d(degrees):
    """
    Format a floating point number as longitude in
    decimal degrees format.

    "East" and "West" added as appropriate

    example::
      In [4]: nucos.format_lon(-33.2)
      Out[4]: '33.2° West'

    :param degrees: degrees of longitude -- negative is South
    :type degrees: float

    :returns: string version of longitude
    """

    return format_latlon2(degrees, LON_POSITIVE_DIRECTION, LON_NEGATIVE_DIRECTION)


def format_lon_dm(degrees):
    """
    Format a floating point number as longitude in
    degrees, decimal minutes format.

    "East" and "West" added as appropriate

    example::
      In [4]: nucos.format_lon(-33.2)
      Out[4]: '33° 12.00′ West'

    :param degrees: degrees of longitude -- negative is South
    :type degrees: float

    :returns: string version of longitude
    """

    return format_latlon2(degrees, LON_POSITIVE_DIRECTION, LON_NEGATIVE_DIRECTION)


def format_lon_dms(f):
    return format_latlon3(f, LON_POSITIVE_DIRECTION, LON_NEGATIVE_DIRECTION)


format_lat = format_lat_dm
format_lon = format_lon_dm



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
tests for unit_conversion main code

designed to be run with pytest

pytest test_unit_conversion.py
"""

import math

try:
    from math import isclose
except ImportError:
    print("nucos > 2.11 only works with Python > 3.7")
    raise

import pytest
from nucos import unit_conversion
from nucos.unit_data import ConvertDataUnits

RELTOL = 1e-5
ABSTOL = 1e-5  # not used, but maybe some day?

KnownValues = [
    # Known values from Handbook of Chemistry and Physics (HCP),
    # except where noted
    ("length", "meters", "feet", 1, 3.2808398),
    ("length", "feet", "meters", 1, .3048),
    ("length", "feet", "miles", 1, 0.000189393),
    ("length", "feet", "nauticalmiles", 1, .00016457883),
    ("length", "feet", "inches", 1, 12),
    ("length", "fathom", "cm", 1, 182.88),
    ("length", "Latitude Minutes", "NauticalMiles", 1.0, 1.0),
    ("Length", "LatitudeDegrees", "NauticalMiles", 1.0, 60),
    ("Length", "micron", "mm", 100, .1),
    ("Length", "\N{MICRO SIGN}m", "m", 1, 1e-6),
    ("Length", "km", "yard", 1, 1093.6133),


    # all values close to value in the "Open Water Oil Identification Job Aid"
    # and close to values in the Unit Conversion sheet distributed with the
    # dispersant mission planner.
    # Technically, oil concentration is a unit of length, but it's conceptually
    # different.
    # So we treat it differently here: (i.e. using bbl/acre as a length
    # would be really weird)
    # Also: we've now got units of mass/area as well -- which is not a length
    ("Oil Concentration", "micron", "mm", 100, .1),
    ("Oil Concentration", "µm", "mm", 1, 1e-3),
    ("Oil Concentration", "in", "mm", 1.0, 25.4),
    ("Oil Concentration", "micron", "bbl/acre", 1.0, 0.02545396),  # calculated from HCP --
    ("Oil Concentration", "bbl/acre", "m^3/km^2", 1.0, 39.2866),  # calculated from HCP --
    ("Oil Concentration", "bbl/acre", "bbl/sq.mile", 1.0, 640.0),  # calculated from HCP --
    ("Oil Concentration", "gal/acre", "bbl/acre", 42.0, 1.0),  # calculated from HCP --
    ("Oil Concentration", "m\N{SUPERSCRIPT THREE}/km\N{SUPERSCRIPT TWO}",
                          "liter/hectare", 1, 10.0),  # calculated from HCP --
    ("Oil Concentration", "l/m\N{SUPERSCRIPT TWO}",
                          "liter/hectare", 1, 10000.0),  # calculated from HCP --

    # These are mass/area -- using an assumed density of 0.95
    ("Oil Concentration", "micron", "g/m^2", 100.0, 95.0),
    ("Oil Concentration", "kg/m^2", "micron", .95, 1000.0),
    ("Oil Concentration", "kg/km^2", "g/m^2", 1.0, 1e-3),
    ("Oil Concentration", "kg/m^2", "g/m^2", 1.0, 1e3),

    ("Area", "sq m", "ft^2", 10, 107.63910),
    ("Area", "Acre", "square yards", 1, 4840),
    ("Area", "Hectares", "cm^2", 1.0, 1e8),
    ("Area", "ha", "sq miles", 1.0, 0.0038610216),
    ("Area", "ft^2", "in^2", 1.0, 144),
    ("Area", "km^2", "ft^2", 1.0, 1.0763910e7),
    ("Area", "nm^2", "acre", 1.0, 847.547741),  # calculated from HCP values
    ("Area", "nm^2", "square mile", 1.0, 1.3242932),  # calculated from HCP values


    ("Volume", "liters", "gal", 1.0, 0.26417205),
    ("Volume", "cubicmeters", "gal", 0.0037854118, 1.0),
    ("Volume", "ml", "cc", 1.0, 1.0),
    ("Volume", "milliongallons", "gal", 1.0, 1e6),
    ("Volume", "liters", "ft\N{SUPERSCRIPT THREE}", 1.0, 0.035314667),
    ("volume", "bbl", "l", 1.0, 158.9873),
    ("volume", "cubicinches", "cubicfeet", 1.0, 0.00057870370),
    ("volume", "cc", "cubicyard", 1.0, 1.3079506e-6),
    ("volume", "fluid ounce (UK)", "fluid oz", 1.0, 0.9607594),
    ("volume", "gallon (UK)", "gal", 1.0, 1.200949),
    ("volume", "cubic kilometer", "m\N{SUPERSCRIPT THREE}", 1.0, 1e9),
    ("volume", "cubic kilometer", "ft^3", 1.0, 3.531467e10),  # the google converter

    ("mass", "kg", "lb", 1.0, 2.2046226),
    ("Mass", "kg", "metrictons", 1.0, 0.001),
    ("mass", "kg", "lb", 1.0, 2.2046226),
    ("mass", "pound", "slug", 1.0, 0.0310810),
    ("mass", "ounce", "lbs", 1.0, 0.0625),
    ("mass", "ton", "gram", 1.0, 907184.74),
    ("mass", "ton(UK)", "ton", 1.0, 1.12),
    ("mass", "mg", "g", 1000.0, 1.0),
    ("mass", "g", "\N{MICRO SIGN}g", 1.0, 1e6),

    ("Time", "seconds", "minutes", 60, 1.0),
    ("time", "days", "minutes", 1.0, 24 * 60),
    ("time", "hr", "seconds", 1.0, 60 * 60),

    ("Velocity", "m/s", "cm/s", 1.0, 100),
    ("Velocity", "m s-1", "cm/s", 1.0, 100),
    ("Velocity", "m.s-1", "cm/s", 1.0, 100),
    ("Velocity", "km/h", "kts", 1.0, 0.5399568),
    ("Velocity", "mph", "ft/s", 1.0, 1.4666666),
    ("Velocity", "ft/min", "mph", 1.0, 0.01136363),
    ("Velocity", "ft/s", "m/min", 1.0, 18.288),
    ("Velocity", "ft/hr", "cm/s", 1.0, 0.0084666),
    ("Velocity", "ft/hr", "km/hr", 1.0, 0.0003048),
    ("Velocity", "knot", "m/s", 1.0, 0.514444),
    ("Velocity", "km/day", "m/s", 1.0, 0.01157404166666666),
    ("Velocity", "km/hr", "km/day", 1.0, 24.0),


    ("Discharge", "cfs", "l/s", 1.0, 28.31685),
    ("Discharge", "gal/hr", "gal/min", 60.0, 1.0),
    ("Discharge", "bbl/day", "l/s", 1.0, (158.9873 / 24 / 3600)),  # calculated from bll=>liter
    ("Discharge", "ft^3/min", "gal/min", 1.0, 7.4805195),
    ("Discharge", "m^3/min", "m^3/s", 1.0, 1.0 / 60.0),
    ("Discharge", "m^3/min", "gal/min", 1.0, 264.1721),
    ("Discharge", "gal/hr", "m^3/min", 1.0, 6.3090197e-5),
    ("Discharge", "gal/hr", "m^3/hr", 1.0, 6.3090197e-5 * 60.0),  # calculated from m^3/min
    ("Discharge", "bbl/day", "gal/day", 1.0, 42),  # from definition of bbl
    ("Discharge", "l/min", "cfs", 1.0, 0.000588578),
    ("Discharge", "bbl/hr", "cfs", 1.0, 5.614583 / 3600.),  # calculated from cfs
    ("Discharge", "cfs", "gal/sec", 1.0, 448.83117 / 60),  # calculated from gal/min

    ("Mass Discharge", "kg/s", "g/s", 1.0, 1000.0),

    ("Kinematic Viscosity", "stokes", "cSt", 1, 100.0),
    ("Kinematic Viscosity", "cm^2/s", "m^2/s", 1, .0001),
    ("Kinematic Viscosity", "cm^2/s", "square inch per second", 1, 0.15500031),
    ("Kinematic Viscosity", "SSU", "cSt", 462.0, 100.0),
    ("Kinematic Viscosity", "SSF", "cSt", 342.0, 724.0),
    ("Kinematic Viscosity", "mm^2/s", "cSt", 1.0, 1.0),  # from ASTM report

    ("Dynamic Viscosity", "kg/(m s)", "Pa s", 1.0, 1.0),
    ("Dynamic Viscosity", "Pa s", "N s/m^2", 1.0, 1.0),
    ("Dynamic Viscosity", "N s/m^2", "kg/(m s)", 1.0, 1.0),
    ("Dynamic Viscosity", "g/(cm s)", "Pa s", 1.0, 0.1),
    ("Dynamic Viscosity", "poise", "Pa s", 1.0, 0.1),
    ("Dynamic Viscosity", "dyne s/cm^2", "Pa s", 1.0, 0.1),
    ("Dynamic Viscosity", "centipoise", "Pa s", 1.0, 0.001),
    ("Dynamic Viscosity", "mPa s", "Pa s", 1.0, 0.001),

    ("temperature", "F", "C", 32, 0),
    ("temperature", "F", "C", 212, 100),
    ("temperature", "C", "K", 0, 273.15),
    ("temperature", "F", "K", 32, 273.15),

    ("deltatemperature", "F", "C", 1.0, (5.0 / 9.0)),
    ("deltatemperature", "F", "C", 0.0, 0.0),
    ("deltatemperature", "C", "K", 12.2, 12.2),
    ("deltatemperature", "K", "F", 10, 90.0 / 5.0),


    ("density", "g/cm^3", "Specific Gravity (15C)", 1, (1 / 0.999016)),
    ("density", "g/cm^3", "Specific Gravity (15C)", 0.999016, 1),
    ("density", "g/L", "kg/m^3", 1.0, 1.0),
    ("density", "kg/L", "kg/m^3", 1.0, 1000.0),
    ("density", "g/mL", "kg/m^3", 1.0, 1000.0),
    ("density", "Specific gravity", "gramspercubiccentimeter", 1, 0.999016),
    ("density", "SG", "API", 1.0, 10),
    ("density", "SG", "API", 2.0, -60.75),
    ("density", "SG", "API", 0.9, 25.7222),
    ("density", "API", "SG", 25.7222, 0.9),
    ("density", "lbs/ft^3", "Grams per Cubic Centimeter", 1.0, 0.016018463),
    ("density", "tonne per cubic meter", "kg/m^3", 1.0, 1000.0),
    ("density", "T/m³", "kg/m^3", 1.0, 1000.0),
    ("density",  "kg/m^3", "pound per gallon", 1.0, 0.0083454),
    ("density",  "lb/gal", "g/cm³", 1.0, 0.119826427),



    # some from https://en.wikipedia.org/wiki/Surface_tension#Physical_units
    ("Interfacial Tension", "dyn/cm", "N/m", 1.0, 0.001),
    ("Interfacial Tension", "N/m", "dyn/cm", 1.0, 1000),
    ("Interfacial Tension", "N/m", "J/m^2", 0.001, 0.001),
    ("Interfacial Tension", "N/m", "mN/m", 1.0, 1000.0),
    ("Interfacial Tension", "N/m", "dyne/cm", 1.0, 1000.0),
    ("Interfacial Tension", "N/m", "pdl/in", 1.0, 0.1837185500656),
    ("Interfacial Tension", "N/m", "lbf/in", 1.0, 0.0057101470975576),
    ("Interfacial Tension", "N/m", "erg/cm^2", 1.0, 1000.0),
    ("Interfacial Tension", "N/m", "erg/cm\N{SUPERSCRIPT TWO}", 1.0, 1000.0),
    ("Interfacial Tension", "N/m", "erg/mm^2", 1.0, 10.0),
    ("Interfacial Tension", "N/m", "erg/mm\N{SUPERSCRIPT TWO}", 1.0, 10.0),

    # ("Adhesion", "Pa", "kPa", 1.0, 0.001),
    # ("Adhesion", "Pa", "MPa", 1.0, 0.000001),
    # ("Adhesion", "Pa", "N/m^2", 1.0, 1.0),
    # ("Adhesion", "Pa", "bar", 1.0, 0.00001),
    # ("Adhesion", "Pa", "mbar", 1.0, 0.01),
    # ("Adhesion", "Pa", "g/cm^2", 1.0, 0.0101972),
    # ("Adhesion", "Pa", "kg/cm^2", 1.0, 0.0000101972),
    # ("Adhesion", "Pa", "dyn/cm^2", 1.0, 10.0),
    # ("Adhesion", "Pa", "psi", 1.0, 0.0001450377),
    # ("Adhesion", "Pa", "lb/in^2", 1.0, 0.0001450377),
    # ("Adhesion", "Pa", "N/m\N{SUPERSCRIPT TWO}", 1.0, 1.0),
    # ("Adhesion", "Pa", "g/cm\N{SUPERSCRIPT TWO}", 1.0, 0.0101972),
    # ("Adhesion", "Pa", "kg/cm\N{SUPERSCRIPT TWO}", 1.0, 0.0000101972),
    # ("Adhesion", "Pa", "dyn/cm\N{SUPERSCRIPT TWO}", 1.0, 10.0),
    # ("Adhesion", "Pa", "lb/in\N{SUPERSCRIPT TWO}", 1.0, 0.0001450377),
    # ("Adhesion", "kg/cm^2", "kg/m^2", 1.0, 10000.0),
    # ("Adhesion", "g/cm^2", "g/m^2", 1.0, 10000.0),
    # ("Adhesion", "kg/cm^2", "g/cm^2", 1.0, 1000.0),
    # ("Adhesion", "kg/m^2", "g/m^2", 1.0, 1000.0),

    ("Concentration In Water", "ppb", "ppm", 1000, 1),  # calculated
    ("Concentration In Water", "fraction", "%", 1, 100),  # calculated
    ("ConcentrationInWater", "lb/ft^3", "mg/l", 1, 16018.450433864),  # hand calculated
    # ("ConcentrationInWater", "mg/l", "lb/ft^3", 160184.50433864002, 1),  # hand calculated
    # ("ConcentrationInWater", "kg/m^3", "lb/ft^3", 16.018464, 1),  # hand calculated
    ("ConcentrationInwater", "mg/l", "ppm", 1.0, 1.0),  # calculated (and kindof defined)
    ("ConcentrationInwater", "mg/l", "ppb", 1.0, 1000),  # calculated
    ("ConcentrationInwater", "mg/kg", "ppb", 1.0, 1000),  # calculated
    ("ConcentrationInWater", "ppt", "percent", 1.0, .1),  # calculated
    ("ConcentrationInWater", "ug/l", "ppb", 1.0, 1.0),  # calculated
    ("ConcentrationInWater", "\N{MICRO SIGN}g/l", "ppb", 1.0, 1.0),  # calculated
    ("ConcentrationInWater", "ug/g", "ppm", 1.0, 1.0),  # calculated
    ("ConcentrationInWater", "mg/ml", "ppm", 1.0, 1000),  # calculated
    ("ConcentrationInWater", "mg/g", "ppt", 1.0, 1.0),  # calculated
    ("ConcentrationInWater", "nanogramperliter", "partpertrillion", 1.0, 1.0),  # calculated
    ("ConcentrationInWater", "g/m\N{SUPERSCRIPT THREE}", "ppm", 1.0, 1.0),  # calculated
    ("ConcentrationInWater", "g/l", "ppm", 1.0, 1000.0),  # calculated
    ("ConcentrationInWater", "kg/l", "part per thousand", 1.0, 1000.0),  # calculated

    ("Concentration", "fraction", "%", 1.0, 100.0),  # calculated
    ("Concentration", "percent", "\N{PER MILLE SIGN}", 1.0, 10.0),  # calculated
    ("Concentration", "\u2030", "ppm", 1.0, 1000.0),  # calculated
    ("Concentration", "ppm", "ppb", 1.0, 1000.0),  # calculated
    ("Concentration", "ppb", "part per trillion", 1.0, 1000.0),  # calculated
    ("Concentration", "parts per trillion", "1", 1.0, 1e-12),  # calculated

    ("Dimensionless", "fraction", "%", 1.0, 100.0),  # calculated
    ("Dimensionless", "percent", "\N{PER MILLE SIGN}", 1.0, 10.0),  # calculated


    ("MassFraction", "fraction", "%", 1, 100),  # calculated
    ("MassFraction", "\N{PER MILLE SIGN}", "percent", 1.0, .1),  # calculated
    ("MassFraction", "ppb", "ppm", 1000, 1),  # calculated
    ("MassFraction", "parts per trillion", "1", 1.0, 1e-12),  # calculated
    ("MassFraction", "g/kg", "fraction", 1.0, 1e-3),  # calculated
    ("MassFraction", "mg/g", "0/00", 1.0, 1.0),  # calculated
    ("MassFraction", "mg/kg", "ppb", 1.0, 1000),  # calculated
    ("MassFraction", "mg/kg", "fraction", 1.0, 1e-6),  # calculated
    ("MassFraction", "ug/g", "ppm", 1.0, 1.0),  # calculated
    ("MassFraction", "ng/g", "fraction", 1.0, 1e-9),  # calculated
    ("MassFraction", "ng/g", "ppb", 1.0, 1.0),  # calculated

    ("VolumeFraction", "fraction", "%", 1, 100),  # calculated
    ("VolumeFraction", "\u2030", "percent", 1.0, .1),  # calculated
    ("VolumeFraction", "ppb", "ppm", 1000, 1),  # calculated
    ("VolumeFraction", "parts per trillion", "1", 1.0, 1e-12),  # calculated
    ("VolumeFraction", "mL/L", "fraction", 1.0, 1e-3),  # calculated
    ("VolumeFraction", "L/m^3", "0/00", 1.0, 1.0),  # calculated

    ("Angular Measure", "degree", "radian", 180.0, math.pi),  # calculated
    ("Angular Measure", "radians", "degrees", 2 * math.pi, 360.0),  # calculated

    ("Angular Velocity", "1/s", "rad/s", 1.0, 1.0),  # definition
    ("Angular Velocity", "hz", "1/s", 5, 5 * 2 * math.pi),  # calculated
    ("Angular Velocity", "rpm", "1/s", 1, 2 * math.pi / 60),  # calculated

    ("Pressure", "Pa", "N/m^2", 1.0, 1.0),  # definition
    ("Pressure", "psi", "Pa", 5.0, 34473.8),  # from google
    ]


def test_new_api_oneshot():
    """
    just to make sure basic API works!

    and these are a few that have caused problems...
    """

    assert isclose(unit_conversion.convert('meter', 'foot', 1),
                   3.28083989501, rel_tol=RELTOL)

    assert isclose(unit_conversion.convert('API', 'SG', 10),
                   1, rel_tol=RELTOL)

    assert isclose(unit_conversion.convert('meter second-1', 'knot', 1),
                   1.94384, rel_tol=RELTOL)

    assert isclose(unit_conversion.convert('m/s', 'knot', 1),
                   1.94384, rel_tol=RELTOL)

    assert isclose(unit_conversion.convert('g/kg', 'ppm', 1),
                   1000, rel_tol=RELTOL)

    assert isclose(unit_conversion.convert('ml/l', '%', 1),
                   0.1, rel_tol=RELTOL)

@pytest.mark.parametrize('unit_type, unit1, unit2, value, new_value',
                         KnownValues)
def test_new_api(unit_type, unit1, unit2, value, new_value):
    """
    this is a parameterized test
    of all the known values, but with the new API
    """
    # filter out the ones that we know are eliminated
    if unit_conversion.Simplify(unit_type) in ('oilconcentration',
                                               'concentrationinwater',
                                               'massfraction',
                                               'volumefraction',
                                               'deltatemperature'):
        return
    assert isclose(unit_conversion.convert(unit1, unit2, value),
                   new_value, rel_tol=RELTOL)


@pytest.mark.parametrize('unit_type, unit1, unit2, value, new_value',
                         KnownValues)
def test_old_api(unit_type, unit1, unit2, value, new_value):
    """
    this is a parameterized test
    of all the known values, with the old API
    """
    # now do the test:
    assert isclose(unit_conversion.convert(unit_type, unit1, unit2, value),
                   new_value, rel_tol=RELTOL)


def test_ConverterClass_init_dupcheck():
    # "name1 is a duplicate -- should be caught"
    unit_dict = {"unit_1": (1.0, ["name1", "name2"]),
                 "unit_2": (3.141592653589793 / 180.0, ["name3", "name1"])
                 }

    with pytest.raises(ValueError):
        unit_conversion.ConverterClass("Random Unit Type", unit_dict)

# bad unit names
def test_bad_type_name():
    with pytest.raises(unit_conversion.InvalidUnitTypeError):
        unit_conversion.convert("BadType", "feet", "miles", 0)

def testBadUnit1():
    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("Length", "eggs", "miles", 0)

def testBadUnit2():
    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("Length", "feet", "spam", 0)

def testBadUnit3():
    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("Density", "API", "feet", 0)


class testOilQuantityConverterClass:
    OQC = unit_conversion.OilQuantityConverter

    def testMassToVolume1(self):
        assert isclose(self.OQC.ToVolume(Mass=1,
                                         MassUnits="metricton",
                                         Density=25,
                                         DensityUnits="API",
                                         VolumeUnits="bbl"),
                       6.9626324,
                       rel_tol=RELTOL)

    def testMassToVolume2(self):
        assert isclose(self.OQC.ToVolume(Mass=1,
                                         MassUnits="metricton",
                                         Density=0.816,
                                         DensityUnits="SG",
                                         VolumeUnits="bbl"),
                       7.71481307, rel_tol=RELTOL)

    def testVolumeToMass1(self):
        Expected = 1.0
        Calculated = self.OQC.ToMass(Volume=6.9626324,
                                     VolUnits="bbls",
                                     Density=25,
                                     DensityUnits="API",
                                     MassUnits="metricton")

        print(Expected, Calculated)
        assert isclose(Expected, Calculated, rel_tol=RELTOL)

    def testVolumeToMass2(self):
        Expected = 1.0
        Calculated = self.OQC.ToMass(Volume=7.83861191,
                                     VolUnits="bbls",
                                     Density=0.816,
                                     DensityUnits="SG",
                                     MassUnits="longton")

        assert isclose(Expected, Calculated, rel_tol=RELTOL)


# fixme: there should probably be a full set of tests for the "new" API
class TestNewConvertAPI():

    def test_bad_convert(self):
        with pytest.raises(unit_conversion.UnitConversionError):
            unit_conversion.convert("kg", "miles", 0)

    def test_invalid_unit(self):
        with pytest.raises(unit_conversion.InvalidUnitError):
            unit_conversion.convert("Mass", "kg", "miles", 0)

    @pytest.mark.parametrize('args',
                             [("kgt", "miles", 0),
                              ("kg", "miless", 0),
                              ("foo", "spam", 0),
                              (0, 0, 0),
                              ])
    def test_NotSupportedUnitError(self, args):
        with pytest.raises(unit_conversion.UnitConversionError):
            unit_conversion.convert(*args)

    @pytest.mark.parametrize('args',
                             [("micron", "g/m^2", 1.0),
                              ])
    def test_invalid_combination(self, args):
        with pytest.raises(unit_conversion.UnitConversionError):
            unit_conversion.convert(*args)


def test_GetUnitTypes():
    # note: not testing all of them
    types = unit_conversion.GetUnitTypes()

    assert "Length" in types
    assert "Temperature" in types
    assert "Area" in types
    assert "Discharge" in types
    assert "Velocity" in types
    assert "Volume" in types


def test_GetUnitNames_length():
    # note: not testing all of them
    #       either all types or all names
    names = unit_conversion.GetUnitNames('Length')
    # print(names)

    assert "meter" in names
    assert "foot" in names
    assert "feet" not in names

def test_GetUnitNames_area():
    names = unit_conversion.GetUnitNames('Area')
    # print(names)

    assert "square foot" in names
    assert "acre" in names
    assert "square meter" in names
    assert "feet" not in names


def test_GetUnitNames_kin_vis():
    names = unit_conversion.GetUnitNames('Kinematic Viscosity')
    print(names)

    assert 'Stoke' in names
    assert 'square centimeter per second' in names


def test_proper_unit_type_names():
    """
    all unit types should be title cased

    This is assumed elsewhere in the code
    """
    for key in ConvertDataUnits.keys():
        assert key == key.title(), f"{key} is not title cased"


def test_GetUnitAbbreviation():
    names = [('Length', 'meter', 'm'),
             ('Volume', 'cubic meter', 'm³'),
             ('Time', 'second', 's'),
             ('Velocity', 'kilometer per hour', 'km/h'),
             ('Discharge', 'cubic foot per second', 'ft³/s'),
             ]

    for unit_type, unit, abrv in names:
        print(unit_type, unit, abrv)
        assert abrv == unit_conversion.GetUnitAbbreviation(unit_type, unit)


def test_FindUnitTypes():
    # just testing that it's there and doesn't crash!
    all_units = unit_conversion.UNIT_TYPES

    assert all_units['s'] == 'time'
    assert all_units['sg'] == 'density'
    assert all_units['feet/s'] == 'velocity'
    assert all_units['m^2/s'] == 'kinematicviscosity'


def test_is_same_unit():
    # a few examples...not complete, but at least it's there and works
    # for some cases

    assert unit_conversion.is_same_unit('knot', 'knots')
    assert unit_conversion.is_same_unit('knot', 'kts')
    assert not unit_conversion.is_same_unit('knot', 'm/s')

    assert unit_conversion.is_same_unit("gal/s", "gal/sec")
    assert unit_conversion.is_same_unit("gallon per second", "gal/sec")

    # these should NOT be the same !
    assert not unit_conversion.is_same_unit("gallon per hour", "gal/sec")

    assert not unit_conversion.is_same_unit("meters", "gal/sec")

    assert not unit_conversion.is_same_unit("something non existant",
                                            "gal/sec")
    assert not unit_conversion.is_same_unit("meter", "meeters")
    assert not unit_conversion.is_same_unit("non_existant", "")


# test the Exceptions
def test_invalid_unit_convert():
    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("length", "flintstones", "meters", 1.0)

    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("length", "feet", "flintstones", 1.0)

    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("temperature", "feet", "C", 1.0)

    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("temperature", "f", "feet", 1.0)

    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("length", "feet", "C", 1.0)

    with pytest.raises(unit_conversion.InvalidUnitTypeError):
        unit_conversion.convert("something_wrong", "feet", "meters", 1.0)

    with pytest.raises(unit_conversion.InvalidUnitError):
        unit_conversion.convert("density", "kg/m^33", "g/cm^3", 1.0)


# these really jsut to get 100% coverage :-)
def test_is_supported():
    """not comprehensive...
    """
    assert unit_conversion.is_supported('foot')
    assert unit_conversion.is_supported('FoOt')
    assert unit_conversion.is_supported('meter second-1')
    assert not unit_conversion.is_supported('something random')


def test_NotSupportedUnitError():
    "just test the __str__..."
    err = str(unit_conversion.NotSupportedUnitError('feeet'))

    assert err == 'The unit: feeet is not supported or not recognized'


def test_InvalidUnitError():
    "just testing the __str__..."
    err = str(unit_conversion.InvalidUnitError(('feet', 'volume')))

    assert err == 'The unit: feet is not in the list for Unit Type: volume'


def test_InvalidUnitTypeError():
    "just test the __str__..."
    err = str(unit_conversion.InvalidUnitTypeError('feeet'))

    assert err == 'The unit type: feeet is not in the UnitConversion database'


class Test_FindAllUnitNames():

    UNIT_NAMES = unit_conversion.FindAllUnitNames()

    def test_length(self):
        names = self.UNIT_NAMES['length']

        assert "meter" in names
        assert "nauticalmiles" in names

        assert "metter" not in names

    def test_dynamic_viscosity(self):
        names = self.UNIT_NAMES['dynamicviscosity']

        print(names)
        assert 'kg/(m s)' in names
        assert 'cP' in names

    def test_get_supported_names(unit_type):
        """
        not the least bit comprehensive, but at shows dict is built
        """
        names = unit_conversion.get_supported_names('Volume')
        names2 = unit_conversion.get_supported_names('volume')

        assert names == names2
        assert 'oz' in names
        assert 'meter' not in names


@pytest.mark.parametrize(("unit_type", "unit"),[('Length', 'meter'),
                                                ('length', 'cM'),
                                                ('oil Concentration', 'micron'),
                                                ])
def test_is_supported_unit(unit_type, unit):
    assert unit_conversion.is_supported_unit(unit_type, unit)


@pytest.mark.parametrize(("unit_type", "unit"),[('Length', 'metter'),
                                                ('length', 'mcM'),
                                                ('oil Concentration', 'sheen'),
                                                ])
def test_is_supported_unit_not(unit_type, unit):
    assert not unit_conversion.is_supported_unit(unit_type, unit)





#!/usr/bin/env python

"""
testing dumping options

not really good tests, but at least shows they run
"""

from __future__ import (unicode_literals,
                        print_function,
                        division,
                        absolute_import
                        )

import os
from nucos import unit_data


def test_write_unit_data_file():
    """
    technically not a test, but a way to get the file re-written automatically
    """
    project_dir = os.path.split(__file__)[0]
    project_dir = os.path.split(project_dir)[0]
    project_dir = os.path.split(project_dir)[0]

    with open(os.path.join(project_dir, "NUCOS_unit_list.rst"), 'w', encoding="utf-8") as outfile:

        outfile.write(
"""
###############
NUCOS Unit List
###############

The NOAA Unit Converter for Oil Spills (NUCOS) is designed specifically to support
oils spill response and planning. As the Oil industry (and the response community)
use some unusual units, this is NOT a general purpose or full featured unit converter.
However, it does try to include all the units that one might need for oil spill work.

Complete unit type, units, and synonym list:

Note that in NUCOS, unit names and synonyms are case and white space insensitive, so, eg:

Pounds per Cubic Foot is the same as poundspercubicfoot

All The Units:
==============
"""
)

        outfile.write(unit_data.all_unit_names("rst"))


def test_dump_to_json():
    '''
    does the dump_to_json_ run without error
    '''
    unit_data.dump_to_json("junk.json")


def test_all_unit_names():
    '''
    only tests if if doesn't crash and give SOME answers
    '''
    names = unit_data.all_unit_names()

    assert "Kinematic Viscosity:" in names
    assert "API degree" in names


def test_write_units():
    """ just makes sure it doesn't crash"""
    unit_data.write_units()
    # and to a file
    unit_data.write_units("all_units.txt")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
unit_data.py

This is the data used by unit_conversion

it also includes a utility function for dumping the units to the console
or a file:
   write_units(filename=None)
"""
from __future__ import unicode_literals, absolute_import
import itertools

ConvertDataUnits = {
    # All lengths in terms of meter
    # All conversion factors from "Handbook of Chemistry and Physics" (HCP)
    # except where noted.

    "Length": {"meter": (1.0, ["m", "meters", "metre"]),
               "centimeter": (0.01, ["cm", "centimeters"]),
               "millimeter": (0.001, ["mm", "millimeters"]),
               "micron": (0.000001, ["\N{MICRO SIGN}m", "micrometer", "microns"]),
               "kilometer": (1000.0, ["km", "kilometers"]),
               "foot": (0.3048, ["ft", "feet"]),
               "inch": (0.0254, ["in", "inches"]),
               "yard": (0.9144, ["yrd", "yards"]),
               "mile": (1609.344, ["mi", "miles"]),
               "nautical mile": (1852.0, ["nm", "nauticalmiles"]),
               "fathom": (1.8288, ["fthm", "fathoms"]),
               "latitude degree": (111120.0, ["latitudedegrees"]),
               "latitude minute": (1852.0, ["latitudeminutes"])
               },

    # this is technically length but used differently, so I'm keeping it
    # separate.
    # micron is the base unit
    "Oil Concentration": {
        "micron": (1.0, ["\N{MICRO SIGN}m", "microns", "micrometer"]),
        "cubic meter per square kilometer": (1.0, ["m\N{SUPERSCRIPT THREE}/km\N{SUPERSCRIPT TWO}", "m^3/km^2"]),
        "millimeter": (1000., ["mm", "millimeters"]),
        "inch": (25400., ["in", "inches"]),
        "barrel per acre": (39.2866176, ["bbl/acre"]),  # calculated from HCP
        "barrel per square mile": (0.06138533995, ["bbl/mile\N{SUPERSCRIPT TWO}",
                                                   "bbl/sq.mile"]),  # calculated from HCP
        "gallon per acre": (0.93539563202687404, ["gal/acre", ]),  # calculated from HCP
        "liter per hectare": (0.1, ["liter/hectare", "l/hectare"]),  # calculated from HCP
        "liter per square meter": (1000.0, ["l/m\N{SUPERSCRIPT TWO}", "l/m^2"]),
        # these are mass/area -- so technically different.
        # but hard coding density of 0.95 -- so we can do it easily.
        "gram per square meter": (1.0526315789473684, ["g/m\N{SUPERSCRIPT TWO}", "g/m^2"]),
        "kilogram per square meter": (1052.6315789473684, ["kg/m\N{SUPERSCRIPT TWO}", "kg/m^2"]),
        "kilogram per square kilometer": (.0010526315789473684, ["g/km\N{SUPERSCRIPT TWO}", "kg/km^2" ]),
    },

    # All Areas in terms of square meter
    "Area": {
        "square meter": (1.0, ["m\N{SUPERSCRIPT TWO}", "m^2", "sq m"]),
        "square centimeter": (.0001, ["cm\N{SUPERSCRIPT TWO}", "cm^2", "sq cm"]),
        "square kilometer": (1e6, ["km\N{SUPERSCRIPT TWO}", "km^2", "sq km"]),
        "acre": (4046.8564, ["ac", "acres"]),
        "square mile": (2589988.1, ["sq miles"]),
        "square nautical mile": (3429904, ["nm\N{SUPERSCRIPT TWO}",
                                           "sq nm", "nm^2"]),
        "square yard": (0.83612736, ["yd\N{SUPERSCRIPT TWO}",
                                     "sq yards", "square yards"]),
        "square foot": (0.09290304, ["ft\N{SUPERSCRIPT TWO}", "ft^2",
                                     "sq foot", "square feet"]),
        "square inch": (0.00064516, ["in\N{SUPERSCRIPT TWO}", "in^2",
                                     "sq inch", "square inches"]),
        "hectare": (10000.0, ["hectares", "ha"]),
    },

    # All volumes in terms of cubic meter
    "Volume": {
        "cubic meter": (1.0, ["m\N{SUPERSCRIPT THREE}", "m^3", "cu m", "cubic meters"]),
        "cubic kilometer": (1e9, ["km\N{SUPERSCRIPT THREE}", "km^3", "cu km", "cubic kilometers"]),
        "cubic centimeter": (1e-6, ["cm\N{SUPERSCRIPT THREE}", "cm^3", "cu cm", "cc"]),
        "milliliter": (1e-6, ["ml", "milliters"]),
        "barrel (petroleum)": (.1589873, ["bbl", "barrels", "barrel", "bbls"]),
        "liter": (1e-3, ["l", "liters"]),
        "gallon": (0.0037854118, ["gal", "gallons", "usgal"]),
        "gallon (UK)": (0.004546090, ["ukgal", "gallons(uk)"]),
        "million US gallon": (3785.4118, ["milliongallons", "milgal"]),
        "cubic foot": (0.028316847, ["ft\N{SUPERSCRIPT THREE}", "ft^3", "cu feet", "cubicfeet"]),
        "cubic inch": (16.387064e-6, ["in\N{SUPERSCRIPT THREE}", "in^3", "cu inch", "cubicinches"]),
        "cubic yard": (.76455486, ["yd\N{SUPERSCRIPT THREE}", "yd^3", "cu yard", "cubicyards"]),
        "fluid ounce": (2.9573530e-5, ["oz", "ounces(fluid)", "fluid oz"]),
        "fluid ounce (UK)": (2.841306e-5, ["ukoz", "fluid oz(uk)"]),
    },

    # All Temperature units in K (multiply by, add)
    "Temperature": {
        "Kelvin": ((1.0, 0.0), ["K", "degrees k", "degree k", "degrees kelvin",
                                "degree kelvin", "deg k"]),
        "Celsius": ((1.0, 273.15), ["C", "degrees c", "degrees celsius",
                                    "deg c", "centigrade"]),
        "Fahrenheit": ((0.55555555555555558, (273.15 * (9. / 5.) - 32.0)),
                       ["F", "degrees f", "degree f", "degrees fahrenheit",
                        "deg f"]),
    },

    # All Temperature units in K (C)
    # This for temperature differences, where you don't want to reset the
    # zero point
    "Delta Temperature": {
        "Kelvin": (1.0, ["K", "degrees k", "degree k", "degrees kelvin",
                         "degree kelvin", "deg k"]),
        "Celsius": (1.0, ["C", "degrees c", "degrees celsius", "deg c",
                          "centigrade"]),
        "Fahrenheit": ((5.0 / 9.0), ["F", "degrees f", "degree f", "deg f",
                                     "degrees fahrenheit"]),
    },


    # All Mass units in Kg (weight is taken to be mass at standard g)
    "Mass": {"kilogram": (1.0, ["kg", "kilograms"]),
             "pound": (0.45359237, ["lb", "pounds", "lbs"]), # not technically mass, but ...
             "gram": (.001, ["g", "grams"]),
             "milligram": (.000001, ["mg"]),
             "microgram": (.000000001, ["\N{MICRO SIGN}g", "ug"]),
             "ton": (907.18474, ["tons", "uston"]),
             "metric ton (tonne)": (1000.0, ["tonnes", "metric ton",
                                             "metric tons", "mt"]),
             "slug": (14.5939, ["slugs"]),
             "ounce": (.028349523, ["oz", "ounces"]),
             "ton (UK)": (1016.0469, ["ukton", "long ton"]),
             },

    # All Time In second
    "Time": {"second": (1.0, ["s", "sec", "seconds"]),
             "minute": (60.0, ["min", "minutes"]),
             "hour": (3600.0, ["hr", "hours", "hrs"]),
             "day": (86400.0, ["days"]),
             },

    # All Velocities in meter per second
    "Velocity": {"meter per second": (1.0, ["m/s", "meter/sec", "meter/second",
                                            "meters per second",
                                            "mps",
                                            "meter sec-1", "meter second-1", "meter seconds-1",
                                            "meters sec-1", "meters second-1", "meters seconds-1",
                                            "m s-1",
                                            ]),
                 "centimeter per second": (.01, ["cm/s"]),
                 "meter per minute": (0.01666666666, ["m/min",
                                                      "meters per minute"]),
                 "kilometer per hour": (0.277777, ["km/h", "km/hr"]),
                 "kilometer per day": (0.0115740416666666, ["km/day", "km/d"]),
                 "knot": (0.514444, ["kts", "knots"]),
                 "mile per hour": (0.44704, ["mph", "miles per hour"]),

                 "foot per second": (0.3048, ["ft/s", "ft/sec",
                                              "feet per second", "feet/s"]),
                 "foot per minute": (0.00508, ["ft/min", "feet per minute",
                                               "feet/min"]),
                 "foot per hour": (0.000084666, ["ft/hr", "feet per hour",
                                                 "feet/hour"]),
                 },

    # All Discharges in cubic meter per second
    "Discharge": {
        "cubic meter per second": (1.0, ["m\N{SUPERSCRIPT THREE}/s", "m^3/s", "cu m/s", "cms"]),
        "cubic meter per min": (1.0 / 60., ["m\N{SUPERSCRIPT THREE}/min", "m^3/min"]),
        "cubic meter per hour": (1.0 / 3600.0, ["m\N{SUPERSCRIPT THREE}/hr", "m^3/hr"]),
        "liter per second": (0.001, ["l/s", "lps"]),
        "liter per minute": (0.001 / 60, ["l/min", ]),
        "cubic foot per second": (.02831685, ["ft\N{SUPERSCRIPT THREE}/s", "cfs", "cu feet/s", "feet^3/s"]),
        "cubic foot per minute": (0.00047194744, ["ft\N{SUPERSCRIPT THREE}/min", "ft^3/min"]),  # calculated from cm^3/s
        "gallon per day": (4.3812636805555563e-08, ["gal/day"]),  # calculated from gal/hr
        "gallon per hour": (1.0515032833333335e-06, ["gal/hr"]),
        "gallon per minute": (6.3090197000000006e-05, ["gal/min", "gpm"]),
        "gallon per second": (0.0037854118, ["gal/s", "gal/sec"]),
        "barrel per hour": (4.4163138888888885e-05, ["bbl/hr"]),
        "barrel per day": (1.84013078e-06, ["bbl/day", "bbl/d"]),  # calculated from bbl/hr
    },

    # All Mass Discharges in kilogram per second
    "Mass Discharge": {
        "kilogram per second": (1.0, ["kg/s"]),
        "gram per second": (0.001, ["g/s"]),
    },


    # Density in g/cc
    # NOTE: Specific Gravity can only be defined for a given
    #       reference temperature.
    #       The most common standard in the oil industry is 15C (or 60F). The
    #       following is based on the value for the Density of water at 15C
    #       (CRC Handbook of Chemistry and Physics)
    "Density": {
        "gram per cubic centimeter": (1.0, ["g/cm\N{SUPERSCRIPT THREE}", "g/cm^3", "grams per cubic centimeter"]),
        "gram per liter": (.001, ["g/L", "gram per litre"]),
        "kilogram per liter": (1.0, ["kg/L", "kilogram per litre"]),
        "gram per milliliter": (1.0, ["g/mL",
                                      "gram per millilitre"]),
        "specific gravity (15\xb0C)": (0.999016, ["S", "specificgravity",  # ASTM D1250
        # "specific gravity (15\xb0C)": (0.99913, ["S", "specificgravity",
                                                 "Spec grav", "SG",
                                                 "specificgravity(15C)"]),
        "kilogram per cubic meter": (.001, ["kg/m\N{SUPERSCRIPT THREE}", "kg/m^3"]),
        "tonne per cubic meter": (1.0, ["tonne/m\N{SUPERSCRIPT THREE}", "tonne/m^3",
                                        "t/m^3", "t/m\N{SUPERSCRIPT THREE}"]),
        "pound per cubic foot": (0.016018463, ["lb/ft\N{SUPERSCRIPT THREE}", "lbs/ft^3", "lb/ft^3"]),
        "pound per gallon": (0.11982643, ["lbs/gal", "lb/gal"]),
        "API degree": (1, ["api"]),  # this is special cased in the code.
    },

    # Kinematic Viscosity in Stokes
    # NOTE: there is a more detailed way to do this, specified in:
    # ASTM D 2161 Standard Practice for Conversion of Kinematic Viscosity to
    # Saybolt Universal Viscosity or to Saybolt Furol Viscosity
    # for the moment, this will only handle approximation for SFS and SUS
    "Kinematic Viscosity": {
        "Stoke": (1.0, ["St", "stokes"]),
        "centiStoke": (.01, ["cSt", "centistokes"]),
        "square millimeter per second": (.01, ["mm\N{SUPERSCRIPT TWO}/s", "mm^2/s"]),
        "square centimeter per second": (1.0, ["cm\N{SUPERSCRIPT TWO}/s",
                                               "cm^2/s"]),
        "square meter per second": (10000, ["m\N{SUPERSCRIPT TWO}/s", "m^2/s"]),
        "square inch per second": (6.4516, ["in\N{SUPERSCRIPT TWO}/s", "in^2/s", "squareinchespersecond"]),
        # from CRC - only good for > 100cSt
        "Saybolt Universal Second": (1 / 462.0, ["SSU", "SUS"]),
        # from Fuel Oil Manual: good for 724cSt
        "Saybolt Furol Second": (0.02116959064, ["SSF", "SFS"]),
    },

    # Dynamic Viscosity
    "Dynamic Viscosity": {
        "kilogram per meter per second": (1.0, ["kg/(m s)"]),
        "Pascal second": (1.0, ["Pa s"]),
        "milliPascal second": (0.001, ["mPa s"]),
        "Newton seconds per square meter": (1.0, ["N s/m\N{SUPERSCRIPT TWO}",
                                                  "N s/m^2",
                                                  ]),

        "gram per centimeter per second": (0.1, ["g/(cm s)"]),
        "poise": (0.1, ["p"]),
        "dyne seconds per square centimeter": (0.1, ["dyne s/cm\N{SUPERSCRIPT TWO}",
                                                     "dyne s/cm^2",
                                                     ]
                                               ),

        "centipoise": (0.001, ["cP"]),
    },

    # Interfacial Tension
    # This is quantified as a force/length measurement in most cases, but a
    # couple exceptions quantify in ergs/area.  An erg is an amount of work,
    # not force, but the conversion is pretty straightforward.
    # FIXME: remove the weirdos: like poundal per inch
    "Interfacial Tension": {
        "Newton per meter": (1.0, ["N/m"]),
        "milliNewton per meter": (0.001, ["mN/m"]),
        "dyne per centimeter": (0.001, ["dyne/cm", "dyn/cm"]),
        "Poundal per inch": (5.443108492, ["pdl/in"]),
        "Pound force per inch": (175.126837, ["lbf/in"]),
        "erg per square centimeter": (0.001, ["erg/cm\N{SUPERSCRIPT TWO}", "erg/cm^2"]),
        "erg per square millimeter": (0.1, ["erg/mm\N{SUPERSCRIPT TWO}", "erg/mm^2"]),
        "joule per square meter": (1.0, ["j/m\N{SUPERSCRIPT TWO}", "j/m^2"]),
    },

    "Pressure": {
        "Pascal": (1.0, ["Pa"]),
        "kiloPascal": (1000.0, ["kPa"]),
        "megaPascal": (1000000.0, ["MPa"]),
        "Newton per square meter": (1.0, ["N/m\N{SUPERSCRIPT TWO}", "N/m^2"]),
        "bar": (100000.0, ["bars"]),
        "millibar": (100.0, ["mbar"]),
        "dyne per square centimeter": (0.1, ["dyn/cm\N{SUPERSCRIPT TWO}", "dyn/cm^2"]),
        "pound per square inch": (6894.76, ["lb/in\N{SUPERSCRIPT TWO}", "lb/in^2", "psi"]),
    },


    # Concentration in water (note: this is converting between mass/volume)
    #                               and mass/mass, hence the "water" part
    "Concentration In Water": {
        "kilogram per cubic meter": (1.0, ["kg/m\N{SUPERSCRIPT THREE}", "kg/m^3"]),
        "gram per cubic meter": (1e-3, ["g/m\N{SUPERSCRIPT THREE}", "g/m^3"]),
        "part per million": (1e-3, ["ppm", "parts per million"]),
        "part per billion": (1e-6, ["ppb", "parts per billion"]),
        "part per thousand": (1.0, [ '\u2030', '0/00', 'ppt', 'parts per thousand']),
        "part per trillion": (1e-9, ["pptr", "parts per trillion"]),
        "fraction (decimal)": (1e3, ["fraction", "mass per mass", "1"]),
        "percent": (10.0, ["%", "parts per hundred"]),
        "pound per cubic foot": (16.018450433864, ["lb/ft\N{SUPERSCRIPT THREE}",
                                                   "lb/ft^3",
                                                   ]),
        "milligram per liter": (0.001, ["mg/l"]),
        "gram per liter": (1.0, ["g/l"]),
        "kilogram per liter": (1000.0, ["kg/l"]),
        "milligram per gram": (1.0, ["mg/g"]),
        "milligram per kilogram": (0.001, ["mg/kg"]),
        "milligram per milliliter": (1.0, ["mg/ml"]),
        "microgram per liter": (1e-6, ["\N{MICRO SIGN}g/l", "ug/l"]),
        "microgram per gram": (1e-3, ["\N{MICRO SIGN}g/g", "ug/g"]),
        "nanogram per liter": (1e-9, ["ng/l"]),
    },

    # Concentration:
    #  This is technically unitless -- just a fraction
    "Concentration": {
        "fraction (decimal)": (1.0, ["fraction", "mass per mass", "1"]),
        "percent": (0.01, ["%", "parts per hundred"]),
        'part per thousand': (1e-3, [ '\u2030', '0/00', 'parts per thousand']),
        "part per million": (1e-6, ["ppm", "parts per million"]),
        "part per billion": (1e-9, ["ppb", "parts per billion"]),
        'part per trillion': (1e-12, ['parts per trillion']),
    },

    "Dimensionless": {
        "fraction (decimal)": (1.0, ["number", "fraction", "1"]),
        "percent": (0.01, ["%", "parts per hundred"]),
        'part per thousand': (1e-3, ['\u2030', '0/00', 'ppt', 'parts per thousand']),
    },

    # Mass Fraction: Any unit that is a mass/mass ratio.  This will be very
    # similar to Concentration, but we want the notion of a generalized mass
    # fraction to be distinct.
    'Mass Fraction': {
        'fraction (decimal)': (1.0, ['fraction', '1', 'mass per mass']),
        'percent': (0.01, ['%', 'parts per hundred']),
        'part per thousand': (1e-3, ['\u2030', '0/00', 'ppt', 'parts per thousand']),
        'part per million': (1e-6, ['ppm', 'parts per million']),
        'part per billion': (1e-9, ['ppb', 'parts per billion']),
        'part per trillion': (1e-12, ['parts per trillion']),
        "gram per kilogram": (1e-3, ["g/kg"]),
        "milligram per gram": (1e-3, ["mg/g"]),
        "milligram per kilogram": (1e-6, ["mg/kg"]),
        "microgram per gram": (1e-6, ["\N{MICRO SIGN}g/g", "ug/g"]),
        "nanogram per gram": (1e-9, ["ng/g", "nanograms per gram"]),
    },

    # Volume Fraction: Any unit that is a volume/volume ratio.  There will be
    # a lot of overlap here with mass fraction, as some units do not allude to
    # a specific unit (% for example).
    'Volume Fraction': {
        'fraction (decimal)': (1.0, ['fraction', '1', 'mass per mass']),
        'percent': (0.01, ['%', 'parts per hundred']),
        'part per thousand': (1e-3, ['\u2030', '0/00', 'ppt', 'parts per thousand']),
        'part per million': (1e-6, ['ppm', 'parts per million']),
        'part per billion': (1e-9, ['ppb', 'parts per billion']),
        'part per trillion': (1e-12, ['parts per trillion']),
        'milliliter per liter': (1e-3, ['ml/l', 'mL/dm^3']),
        'liter per cubic meter': (1e-3, ['l/m^3']),
    },

    "Angular Measure": {
        "radians": (1.0, ["radian", "rad"]),
        "degrees": (3.141592653589793 / 180.0, ["degree", "deg"])
    },

    "Angular Velocity": {
        "rad/s": (1.0, ["1/s", "radians/sec"]),
        "hertz": (2 * 3.141592653589793, ["hz", "cycles/sec"]),
        "rpm": (3.141592653589793 / 30, ["rotations per minute"])
    },

}


#fixme -- this was already here, yes ???
# need to check for duplicate (and correct) functionality)
# Build the unit sets to allow quick lookup of type and conversion legality
# this creates something like the following:
# unit_sets = {'Length': set(['m','km','mm',...]),
#              'Area': set(['m^2','cm^2',...]),
#              ...
#              }

unit_sets = {}
for k in ConvertDataUnits.keys():
    unit_sets[k] = set(itertools.chain(*[y for (x, y)
                                         in ConvertDataUnits[k].values()]))

# fixme: what is this used for?
# if it's just for listing known units, then we should leave it in
# del unit_sets['Concentration In Water']

# not always length -- can be mass/area too.
# unit_sets['Oil Concentration'] = unit_sets['Length']

# Build the global unit list

supported_units = set([])
for s in unit_sets.values():
    supported_units = supported_units.union(s)


def write_units(filename=None):
    """
    write all the units out to a file.

    :param filename=None: filename to write to. If None, it will write to stdout.

    """
    import sys
    if filename is None:
        f = sys.stdout
    else:
        f = open(filename, 'w', encoding="utf-8")
    f.write("NUCOS unit set:\n")
    for key, value in ConvertDataUnits.items():
        f.write("\n%s:\n" % key)
        for key2 in value:
            f.write("    %s\n" % key2)


def all_unit_names(format="str"):
    """
    returns a string of all unit names, grouped by unit type

    :param format="str": format for output -- default is plain python str
                         other option: "rst" for restuctured text
    """
    if format == "str":
        result = []
        for key, value in ConvertDataUnits.items():
            result.append('\n%s:\n' % key)
            for key2 in value:
                result.append("    %s\n        " % key2)
                result.append(", ".join(value[key2][1]))
                result.append("\n")
        return "".join(result)
    elif format == "rst":
        result = []
        for unit_type, value in ConvertDataUnits.items():
            result.append('\n%s:\n' % unit_type)
            result.append('-' * (len(unit_type) + 1) + '\n\n')
            for unit in value:
                result.append("%s\n" % unit)
                result.append('.' * len(unit) + '\n\n    ')
                result.append(", ".join(value[unit][1]))
                result.append("\n\n")
        return "".join(result)
    else:
        raise ValueError('only supported formats are "str" and "rst"')


def dump_to_json(filename=None):
    """
    dumps the full unit data to JSON, for use in the Javascript version, or ...
    """
    import sys
    import json

    f = open(filename, 'w') if filename else sys.stdout
    f.write(json.dumps(ConvertDataUnits, indent=2, separators=(',', ':')))



#!/usr/bin/env python

"""
unit_conversion package

All it really contains is the one unit_conversion module,
a data module, plus a module for lat-lon conversion.

All of unit_conversion is imported here for convenience

"""

__version__ = "3.1.2"


from .unit_conversion import (UnitConversionError,
                              InvalidUnitError,
                              InvalidUnitTypeError,
                              NotSupportedUnitError,
                              is_same_unit,
                              is_supported,
                              is_supported_unit,
                              get_supported_names,
                              # not sure these should be used externally
                              FindUnitTypes,
                              GetUnitTypes,
                              GetUnitNames,
                              GetUnitAbbreviation,
                              # ConverterClass,
                              # Converters,
                              # UNIT_TYPES,
                              # Simplify,
                              # DensityConverterClass,
                              # TempConverterClass,
                              # LatLongConverter,
                              # OilQuantityConverter,
                              convert,
                              )

from .lat_long import (LatLongConverter,
                       format_lat,
                       format_lon,
                       format_lat_d,
                       format_lon_d,
                       format_lat_dm,
                       format_lon_dm,
                       format_lat_dms,
                       format_lon_dms,
                       )

# this should probably not be exposed
from .unit_data import ConvertDataUnits

from nucos import lat_long




#!/usr/bin/env python

from nucos import lat_long
import unittest
import pytest


class testSignBit(unittest.TestCase):
    
    def testNeg(self):
        self.assertTrue(lat_long.signbit(-5.0))
    def testPos(self):
        self.assertFalse(lat_long.signbit(5.0))
    def testNegZero(self):
        self.assertTrue(lat_long.signbit(-0.0))
    def testPosZero(self):
        self.assertFalse(lat_long.signbit(0.0))
    def testIntegerNeg(self):
        self.assertTrue(lat_long.signbit(-5))
    def testIntegerPos(self):
        self.assertFalse(lat_long.signbit(5))


class testLatLongErrors(unittest.TestCase):
    LLC = lat_long.LatLongConverter

    def testNegativeMinutes(self):
        self.assertRaises(ValueError, self.LLC.ToDecDeg, 30, -30)

    def testNegativeSeconds(self):
        self.assertRaises(ValueError, self.LLC.ToDecDeg, 30, 30, -1)

    def testTooBig(self):
        self.assertRaises(ValueError , self.LLC.ToDecDeg, d=200)

    def testTooNegative(self):
        self.assertRaises(ValueError , self.LLC.ToDecDeg, d=-181)

    def testTooBigMin(self):
        self.assertRaises(ValueError , self.LLC.ToDecDeg, d=20, m=61)

    def testTooBigSec(self):
        self.assertRaises(ValueError , self.LLC.ToDecDeg, d=30, m=42, s=61)

    def testDegFractAndMin(self):
        self.assertRaises(ValueError , self.LLC.ToDecDeg, d=30.2, m=5, s=0)

    def testDegFractAndSec(self):
        self.assertRaises(ValueError , self.LLC.ToDecDeg, d=30.2, m=0, s=6.3)
        
    def testMinFractAndSec(self):
        self.assertRaises(ValueError , self.LLC.ToDecDeg, d=30, m=4.5, s=6)
        

class testLatLong(unittest.TestCase):
    LLC = lat_long.LatLongConverter

    def testDecDegrees(self):
        self.assertEqual(self.LLC.ToDecDeg(30, 30), 30.5)

    def testDecDegrees2(self):
        self.assertAlmostEqual(self.LLC.ToDecDeg(30, 30, 30), 30.50833333333)

    def testDegMin(self):
        self.assertEqual(self.LLC.ToDegMin(30.5)[0], 30 )
        self.assertAlmostEqual(self.LLC.ToDegMin(30.5)[1], 30.0 )
   
    def testMinusZeroDeg(self):
        self.assertEqual(self.LLC.ToDecDeg(d=-0.0, m=20, s=20), -0.33888888888888885)

    def testBinaryProblem(self):
        self.assertEqual(self.LLC.ToDegMinSec(45.05), (45, 3, 0.0))
        
    def testDDString(self):
        d, m, s = 120, 30, 5
        DecDeg = self.LLC.ToDecDeg(d, m, s, ustring=True)
        self.assertEqual(self.LLC.ToDecDeg(d, m, s, ustring=True),
                         u"120.501389\xb0")

    def testDDString2(self):
        d, m, s = -50, 30, 5
        DecDeg = self.LLC.ToDecDeg(d, m, s, ustring=True)
        self.assertEqual(self.LLC.ToDecDeg(d, m, s, ustring=True),
                         u"-50.501389\xb0")

    def testDDString2(self):
        d, m, s = 0, 30, 0
        DecDeg = self.LLC.ToDecDeg(d, m, s, ustring=True)
        self.assertEqual(self.LLC.ToDecDeg(d, m, s, ustring=True),
                         u"0.500000\xb0")

    def testDMString(self):
        d, m = 120, 45.5
        DecDeg = self.LLC.ToDecDeg(d, m)
        self.assertEqual(self.LLC.ToDegMin(DecDeg, True),
                         u"120\xb0 45.500'")

    def testDMString2(self):
        d, m = -120, 3
        DecDeg = self.LLC.ToDecDeg(d, m)
        self.assertEqual(self.LLC.ToDegMin(DecDeg, True),
                         u"-120\xb0 3.000'")

    def testDMSString(self):
        d, m, s = 120, 45,  15
        DecDeg = self.LLC.ToDecDeg(d, m, s)
        self.assertEqual(self.LLC.ToDegMinSec(DecDeg, True),
                         u"120\xb0 45' 15.00\"")

    def testDMSString2(self):
        d, m, s = -120, 3,  15
        DecDeg = self.LLC.ToDecDeg(d, m, s)
        self.assertEqual(self.LLC.ToDegMinSec(DecDeg, True),
                         u"-120\xb0 3' 15.00\"")

    def testDMtringZero(self):
        d, m, s = -0.0, 3,  0
        DecDeg = self.LLC.ToDecDeg(d, m, s)
        self.assertEqual(self.LLC.ToDegMin(DecDeg, True),
                         u"""-0\xb0 3.000'""")

    def testDMSStringZero(self):
        d, m, s = -0.0, 3,  15
        DecDeg = self.LLC.ToDecDeg(d, m, s)
        self.assertEqual(self.LLC.ToDegMinSec(DecDeg, True),
                         u'''-0\xb0 3' 15.00"''')



class testLatitude(unittest.TestCase):
    L = lat_long.Latitude

    def testTooBig(self):
        self.assertRaises(ValueError , self.L, deg=95)

    def testTooNeg(self):
        self.assertRaises(ValueError , self.L, deg=-95)

    def testSignAndDir(self):
        self.assertRaises(ValueError , self.L, deg=-45, direction="N")

    def testNegativeSeconds(self):
        self.assertRaises(ValueError, self.L, deg=0.0, min=0.0, sec=-0.001)

    def testNegativeMinutes(self):
        self.assertRaises(ValueError, self.L, deg=0.0, min=-0.1)

    def testDirS(self):
        self.assertEqual(self.L(deg=45, direction="S").value, -45.0)
        
    def testDirN(self):
        self.assertEqual(self.L(deg=45.5, direction="N").value, 45.5)

    def testDirWrong(self):
        self.assertRaises(ValueError, self.L, deg=45, direction="Whatever")
        
    def testDM1(self):
        self.assertEqual(self.L(deg=45.5, direction="S").degrees_minutes(), (45, 30.0, 'South'))

    def testDMS1(self):
        self.assertEqual(self.L(deg=45.5, direction="S").degrees_minutes_seconds(), (45, 30.0, 0, 'South'))

    def testDMS2(self):
        self.assertEqual(self.L(deg=30.001).degrees_minutes_seconds(), (30, 0, 3.6, 'North'))
        
    def testFormat1(self):
        self.assertEqual(self.L(deg=30.001).format(1), u'30.00\xb0 North')
        
    def testFormat2(self):
        self.assertEqual(self.L(deg=30.001).format(2), u"30\xb0 0.06' North")
        
    def testFormat3(self):
        self.assertEqual(self.L(deg=30.001).format(3), u'''30\xb0 0' 3.60" North''')
        
    def testFormat4(self):
        self.assertEqual(self.L(deg=-30.001).format(1), u'30.00\xb0 South')
        
    def testFormatHTML3(self):
        self.assertEqual(self.L(deg=30.001).format_html(3), '30&deg; 0\' 3.60" North')

    def testDirectionN(self):
        self.assertEqual(self.L(deg=80.5).direction(), 'North')

    def testDirectionS(self):
        self.assertEqual(self.L(deg=-80.5).direction(), 'South')

    def testDirectionZero(self):
        # note: should this be North or South???
        self.assertEqual(self.L(deg=0.0).direction(), 'North')

    def testDirectionSmall(self):
        self.assertEqual(self.L(deg=-.00000000000001).direction(), 'South')

    
class testLongitude(unittest.TestCase):
    L = lat_long.Longitude

    def testTooBig(self):
        self.assertRaises(ValueError, self.L, deg=185)

    def testTooNeg(self):
        self.assertRaises(ValueError, self.L, deg=-185)

    def testSignAndDir(self):
        self.assertRaises(ValueError, self.L, deg=-45, direction="W")

    def testDirW(self):
        self.assertEqual(self.L(deg=45, direction="W").value, -45.0)

    def testDirE(self):
        self.assertEqual(self.L(deg=45.5, direction="E").value, 45.5)

    def testDirWrong(self):
        self.assertRaises(ValueError, self.L, deg=45, direction="Anything")
        
    def testDM1(self):
        self.assertEqual(self.L(deg=45.5, direction="W").degrees_minutes(), (45, 30.0, 'West'))

    def testDMS1(self):
        self.assertEqual(self.L(deg=45.5, direction="E").degrees_minutes_seconds(), (45, 30.0, 0, 'East'))

    def testDMS2(self):
        self.assertEqual(self.L(deg=30.001).degrees_minutes_seconds(), (30, 0, 3.6, 'East'))
        
    def testFormat1(self):
        self.assertEqual(self.L(deg=30.001).format(1), u'30.00\xb0 East')
        
    def testFormat2(self):
        self.assertEqual(self.L(deg=30.001).format(2), u"30\xb0 0.06' East")
        
    def testFormat3(self):
        self.assertEqual(self.L(deg=30.001).format(3), u'''30\xb0 0' 3.60" East''')
        
    def testFormat4(self):
        self.assertEqual(self.L(deg=-30.001).format(1), u'30.00\xb0 West')
        
    def testFormatHTML3(self):
        self.assertEqual(self.L(deg=30.001).format_html(3), '30&deg; 0\' 3.60" East')

    def testDirectionN(self):
        self.assertEqual(self.L(deg=80.5).direction(), 'East')

    def testDirectionS(self):
        self.assertEqual(self.L(deg=-80.5).direction(), 'West')

    def testDirectionZero(self):
        # note: should this be East or West???
        self.assertEqual(self.L(deg=0.0).direction(), 'East')

    def testDirectionSmall(self):
        self.assertEqual(self.L(deg=-.00000000000001).direction(), 'West')


# tests the format* functions -- used by ResponseLink

@pytest.mark.parametrize(("number", "text"),
                         [(28.2186111111, "28\xb0 13.12\u2032 North"),  # from incident 6652
                          (-0.1, "0\xb0 6.00\u2032 South")
                         ])
def test_format_lat(number, text):
    assert lat_long.format_lat(number) == text


@pytest.mark.parametrize(("number", "text"),
                         [(-92.6244444444, "92\xb0 37.47\u2032 West"),  # from incident 6652
                         ])
def test_format_lon(number, text):
    assert lat_long.format_lon(number) == text


## degrees, minutes, seconds formatting:
@pytest.mark.parametrize(("number", "text"),
                         [(28.2186111111, "28\xb0 13\u2032 7.00\u2033 North"), #from incident 6652
                          (-0.1, "0\xb0 6\u2032 0.00\u2033 South")
                         ])
def test_format_lat_dms(number, text):
    assert lat_long.format_lat_dms(number) == text

@pytest.mark.parametrize(("number", "text"),
                         [(-92.6244444444, "92\xb0 37\u2032 28.00\u2033 West"), #from incident 6652
                         ])
def test_format_lon_dms(number, text):
    assert lat_long.format_lon_dms(number) == text




"""
a couple tests to make sure the old package name still works

These just happen to be the ones used on ResponseLink :-)
"""

import nucos
from math import isclose
import unit_conversion

def test_version():
    assert unit_conversion.__version__ == nucos.__version__


def test_convert():
    result = unit_conversion.convert("Volume", 'liter', "gallons", 10.0)

    assert isclose(result, 2.6417205, rel_tol=1e-6)


def test_lat_format():
    assert unit_conversion.lat_long.format_lat(-32.5) == '32° 30.00′ South'


def test_lon_format():
    assert unit_conversion.lat_long.format_lon(-32.5) == '32° 30.00′ West'






"""
This here for backward compatibility only

It intends to present all the names actually used in the old
unit_conversion namespace

This API is deprecated -- please use the nucos package instead.
"""

import warnings

from nucos import __version__

from nucos.unit_conversion import *

from nucos import lat_long

from nucos.lat_long import (LatLongConverter,
                            Latitude,
                            Longitude,
                            format_lat,
                            format_lon,
                            format_lat_d,
                            format_lon_d,
                            format_lat_dm,
                            format_lon_dm,
                            format_lat_dms,
                            format_lon_dms,
                            )

warnings.warn(DeprecationWarning(
    "The unit_conversion module is deprecated -- "
    "import as `nucos` instead"))


